from flask import Blueprint, request
from flask_restful import Api, Resource, abort
from marshmallow import ValidationError

from portal.rest import GroupSchema, UserSchema
from portal.service import service
from portal.tools.logging import log
from portal.database.models import Group

groups = Blueprint('groups', __name__, url_prefix='/courses/<string:cid>/groups')
groups_api = Api(groups)

group_schema = GroupSchema()
groups_schema = GroupSchema(many=True, only=('id', 'name'))
user_schema = UserSchema()
users_schema = UserSchema(many=True, only=('id', 'username', 'uco', 'email'))


class GroupResource(Resource):
    def get(self, cid, gid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not find group {gid} in course {cid}: course not found.")
        group = service.find_group(course, gid)
        if not group:
            abort(404, message=f"Could not find group {gid} in course {cid}: group not found.")
        return group_schema.dump(group)

    def delete(self, cid, gid):
        log.info(f"DELETE to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not delete group {gid} in course {cid}: course not found.")
        group = service.find_group(course, gid)
        if not group:
            abort(404, message=f"Could not delete group {gid} in course {cid}: group not found.")
        service.delete_entity(group)
        return '', 204

    def put(self, cid, gid):
        log.info(f"PUT to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not update group {gid} in course {cid}: course not found.")
        group = service.find_group(course, gid)
        if not group:
            abort(404, message=f"Could not update group {gid} in course {cid}: group not found.")

        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for group update.')
        try:
            data = group_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        group.name = data['name']
        service.write_entity(group)
        log.debug(f"Updated group {gid} to {group}.")
        return '', 204


class GroupsList(Resource):
    def get(self, cid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not list groups in course {cid}: course not found.")

        return groups_schema.dump(course.groups)

    def post(self, cid):
        log.info(f"POST to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not create group in course {cid}: course not found.")
        json_data = request.get_json()
        if not json_data:
            abort(400, message='No data provided for group creation.')
        try:
            data = group_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        new_group = Group(course=course, name=data['name'])
        service.write_entity(new_group)
        log.info(f"Created group={new_group} for course {cid}")
        return group_schema.dump(new_group)[0], 201


class GroupUsersList(Resource):
    def get(self, cid, gid):
        log.info(f"GET to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not list users in group {gid} in course {cid}: course not found.")
        group = service.find_group(course, gid)
        if not group:
            abort(404, message=f"Could not list users in group {gid} in course {cid}: group not found.")

        return users_schema.dump(group.users)

    def put(self, cid, gid):
        log.info(f"PUT to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not update users of group {gid} in course {cid}: course not found.")
        group = service.find_group(course, gid)
        if not group:
            abort(404, message=f"Could not update users of group {gid} in course {cid}: group not found.")

        json_data = request.get_json()
        try:
            data = users_schema.load(json_data)[0]
        except ValidationError as err:
            log.info(f"Validation failed on: {err.messages}")
            return abort(422, errors=err.messages)  # TODO: flash, process errors

        # everything from users_add is added, THEN everything from users_remove is subtracted
        users_old = set(group.users)
        users_add = data.get('add')
        users_remove = data.get('remove')
        if users_add is not None:
            users_old.union(users_add)
        if users_remove is not None:
            users_old.difference(users_remove)

        log.info(f"Updated users of group {gid} to {data}.")
        return '', 204


class GroupUser(Resource):
    def put(self, cid, gid, uid):
        log.info(f"PUT to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not add user {uid} to group {gid} in course {cid}: course not found.")
        group = service.find_group(course, gid)
        if not group:
            abort(404, message=f"Could not add user {uid} to group {gid} in course {cid}: group not found.")
        user = service.find_user(uid)
        if not user:
            abort(404, message=f"Could not add user {uid} to group {gid} in course {cid}: user not found.")

        if user not in group.users:
            group.users.append(user)
            log.info(f"Added user {uid} to group {gid} in course {cid}.")
        else:
            log.info(f"User {uid} is already in group {gid} in course {cid}: no change.")
        return '', 204

    def delete(self, cid, gid, uid):
        log.info(f"DELETE to {request.url}")
        course = service.find_course(cid)
        if not course:
            abort(404, message=f"Could not remove user {uid} from group {gid} in course {cid}: course not found.")
        group = service.find_group(course, gid)
        if not group:
            abort(404, message=f"Could not remove user {uid} from group {gid} in course {cid}: group not found.")
        user = service.find_user(uid)
        if not user:
            abort(404, message=f"Could not remove user {uid} from group {gid} in course {cid}: user not found.")

        try:
            group.users.remove(user)
        except ValueError:
            abort(400, message=f"Could not remove user {uid} from group {gid} in course {cid}: "
                               f"group does not contain user.")
        log.info(f"Removed user {uid} from group {gid} in course {cid}.")
        return '', 204


groups_api.add_resource(GroupResource, '/<string:gid>')
groups_api.add_resource(GroupsList, '/')
groups_api.add_resource(GroupUsersList, '/<string:gid>/users')
groups_api.add_resource(GroupUser, '/<string:gid>/users/<string:uid>')
