from portal.tools.logging import log
from portal import db
from portal.database.models import Course, Role, User, Project, Group, Submission, SubmissionState
from portal import storage


def delete_entity(entity):
    db.session.remove(entity)
    db.session.commit()


def write_entity(entity):
    db.session.add(entity)
    db.session.commit()


def find_submission(identifier):
    log.debug(f"Attempting submission search by id.")
    submission = Submission.query.filter_by(id=identifier).first()
    if submission:
        return submission

    log.debug(f"Did not find submission with id={identifier}.")
    return None


def find_course(identifier):
    log.debug(f"Attempting course search by id.")
    course = Course.query.filter_by(id=identifier).first()
    if course:
        return course

    log.debug(f"Did not find course with id={identifier}.")
    log.debug(f"Attempting course search by codename.")
    course = Course.query.filter_by(codename=identifier).first()
    if course:
        return course

    log.debug(f"Did not find course with codename={identifier}.")
    return None


def find_project(course, identifier):
    log.debug(f"Attempting project search by id.")
    project = Project.query.filter_by(course=course).filter_by(id=identifier).first()
    if project:
        return project

    log.debug(f"Did not find project with id={identifier}.")
    log.debug(f"Attempting project search by name.")
    project = Project.query.filter_by(course=course).filter_by(name=identifier).first()
    if project:
        return project

    log.debug(f"Did not find project with codename={identifier}.")
    return None


def find_group(course, identifier):
    log.debug(f"Attempting group search by id.")
    group = Group.query.filter_by(course=course).filter_by(id=identifier).first()
    if group:
        return group

    log.debug(f"Did not find group with id={identifier}.")
    log.debug(f"Attempting group search by name.")
    group = Group.query.filter_by(course=course).filter_by(name=identifier).first()
    if group:
        return group

    log.debug(f"Did not find group with name={identifier}.")
    return None


def find_role(course, identifier):
    log.debug(f"Attempting role search by id.")
    role = Role.query.filter_by(course=course).filter_by(id=identifier).first()
    if role:
        return role

    log.debug(f"Did not find role with id={identifier} in course {course.codename}.")
    log.debug(f"Attempting role search by name.")
    role = Role.query.filter_by(course=course).filter_by(name=identifier).first()
    if role:
        return role

    log.debug(f"Did not find role with name={identifier} in course {course.codename}.")
    return None


def find_user(identifier):
    log.debug(f"Attempting user search by id.")
    user = User.query.filter_by(id=identifier).first()
    if user:
        return user

    log.debug(f"Did not find user with id={identifier}.")
    log.debug(f"Attempting user search by email.")

    user = User.query.filter_by(email=identifier).first()
    if user:
        return user
    log.debug(f"Did not find user with email={identifier}.")
    log.debug(f"Attempting user search by username.")

    user = User.query.filter_by(username=identifier).first()
    if user:
        return user

    log.debug(f"Did not find user with username={identifier}.")
    # not searching by uco
    return None


def can_create_submission(user, project) -> bool:
    """Checks if a user may create a new submission.

    A user is permitted to create a new submission if there are no submissions
    created or queued for him in the given project.

    Args:
        user(User): the user who wants to crete a new submission
        project(Project): the project in which the user wants to create a submission

    Returns: whether the user can create a new submission in the project

    """
    conflict = Submission.query.filter(Submission.user == user)\
        .filter(Submission.project == project)\
        .filter(Submission.state.in_([SubmissionState.CREATED, SubmissionState.READY, SubmissionState.QUEUED])).first()

    return conflict is None


def create_submission(user, project, file_params, project_params_string) -> Submission:
    """Creates a new submission in the database and downloads its files.

    Zip and git sources are processed differently.

    Args:
        user(User): the submission's author
        project(Project): the project the submission falls under
        file_params(dict): parameters for Storage - how and where to download files
        project_params_string(str): custom parameters specified by project with their values as a JSON-encoded string

    Returns(Submission): the created submission entity

    """
    # adding review at submission create is not supported
    # TODO: store all configuration relevant for submission scheduling:
    # project.submissions_allowed from, to, archive; project.config.submission_scheduler_config, submission_params?
    new_submission = Submission(user=user, project=project, parameters=project_params_string)
    write_entity(new_submission)

    # download relevant files into Storage
    file_params['whitelist'] = project.config.file_whitelist
    if file_params['source']['type'] == 'zip':
        file_params['source']['path'] = "todo"  # TODO: find out how to download files through flask
    storage.submissions.create(new_submission.id, **file_params)
    return new_submission
